@@ -1,55 +1,59 @@
-'''
+# -*- encoding: utf-8 -*-
+
+"""
 Created on Jun 14, 2012
 
 @author: Steve Ivy <steveivy@gmail.com>
 @co-author: yangming <yangming@appfirst.com>
 http://www.appfirst.com
 
-python client for appfirst statsd+
-this file expects local_settings.py to be in the same dir, with statsd host and port information:
+Updated for Python 3 May 14, 2014 by michael@appfirst.com
+
+Python client for AppFirst Statsd+
+this file expects local_settings.py to be in the same dir, with statsd host and
+port information:
 
 statsd_host = 'localhost'
 statsd_port = 8125
 
-Sends statistics to the stats daemon over UDP
-Sends statistics to the appfirst collector over UDP
-'''
+Sends statistics to a stats daemon over UDP
+Sends statistics to the AppFirst collector over POSIX MQ or Windows Mailslot
+"""
+
 import sys
 import time
 import random
+import atexit
 import threading
 from socket import socket, AF_INET, SOCK_DGRAM
 
+
 #---------------------------------------------------------------------------
 #   Default UDP Transport
 #---------------------------------------------------------------------------
 class UDPTransport(object):
+    def __init__(self, host='localhost', port=8125):
+        self.host = host
+        self.port = port
+
     def emit(self, data):
         """
-        Squirt the metrics over UDP
+        Send the metrics over UDP
         """
-        try:
-            import local_settings as settings
-            host = settings.statsd_host
-            port = settings.statsd_port
-            addr=(host, port)
-        except Exception:
-            exit(1)
+        addr=(self.host, self.port)
 
         udp_sock = socket(AF_INET, SOCK_DGRAM)
         try:
-            for stat in data.keys():
-                value = data[stat]
-                send_data = "%s:%s" % (stat, value)
+            for name, value in data.items():
+                send_data = "{0}:{1}".format(name, value.format_string(udp=True))
                 udp_sock.sendto(send_data, addr)
-        except:
-            from pprint import pprint
-            print "Unexpected error:", pprint(sys.exc_info())
-            pass # we don't care
+        except Exception as e:
+            sys.stderr.write("Error emitting stats over UDP: {0.__class__.__name__}: {0}\n".format(e))
 
     def close(self):
         pass
 
+
 #---------------------------------------------------------------------------
 #   Statsd Aggregator to buffer stats of the same bucket and dump them together
 #---------------------------------------------------------------------------
@@ -61,23 +65,23 @@ class StatsdAggregator(object):
         self.buf = {}
         self.lock = threading.Lock()
         self._service_thread = None
-        
+
         self.left_buffers = {}   # 2 buffer groups, each stored in a dict
         self.right_buffers = {}  # one of each for each thread
         self.rbufs = self.left_buffers    # buffer group currently being read from
         self.wbufs = self.right_buffers   # buffer group currently being written to
-    
+
     def service_loop(self):
-        #print "launching service loop"
         while self.running:
-            #print "top of service loop"
             time.sleep(self.interval/2.0)
-            self.swap_buffers()        
+            self.swap_buffers()
             time.sleep(self.interval/2.0)
             self.dump()
 
     def start(self):
-        #print "starting aggregation"
+        """
+        Start aggregation
+        """
         if self.running:
             return
         else:
@@ -86,16 +90,21 @@ class StatsdAggregator(object):
                 self._service_thread = threading.Thread(target=self.service_loop)
                 self._service_thread.daemon = True
             self._service_thread.start()
-    
+
     def stop(self):
-        #print "stopping aggregation"
+        """
+        Stop aggregation
+        """
         if self.running:
-            self.running = False    
+            self.running = False
             self.dump()
             self.swap_buffers()
             self.dump()
-        
+
     def is_empty(self):
+        """
+        Check if data in self.buf
+        """
         if self.buf:
             return False
         else:
@@ -104,7 +113,7 @@ class StatsdAggregator(object):
     def add(self, bucket):
         # is setdefault atomic (thread safe)?  It's faster!
         write_buffer = self.wbufs.setdefault(threading.currentThread(), {})
-        '''
+        """
         if threading.currentThread() in self.wbufs:
             write_buffer = self.wbufs[threading.currentThread()]
         else:
@@ -112,33 +121,31 @@ class StatsdAggregator(object):
             write_buffer = {}
             self.lock.acquire()
             self.wbufs[threading.currentThread()] = write_buffer
-            self.lock.release()    
-        '''    
+            self.lock.release()
+        """
         if bucket.name in write_buffer:
-            #print "\texisting bucket"
+            # aggregate if bucket is already in bucket
             write_buffer[bucket.name].aggregate(bucket.stat)
         else:
-            #print "\tvirgin bucket"
+            # otherwise add
             write_buffer[bucket.name] = bucket
-        return 
+        return
 
     def dump(self):
-        # aggregate data across all read buffers
+        """
+        aggregate data across all read buffers and then emit
+        """
         send_buffer = {}
         for th in self.rbufs:
-            #print "dump:found an rbuf"
-            read_buffer = self.rbufs[th]       
-            for name, bucket in read_buffer.iteritems():
-                #print "dump:found a bucket"
+            read_buffer = self.rbufs[th]
+            for name, bucket in read_buffer.items():
                 if name in send_buffer:
-                    #print "dump: existsing agg bucket"
                     send_buffer[name].aggregate(bucket.stat)
                 else:
-                    #print "dump: virgin agg bucket"
                     send_buffer[name]=bucket
-            read_buffer.clear()   
+            read_buffer.clear()
         self.transport.emit(send_buffer)
-        
+
     def swap_buffers(self):
         if self.rbufs == self.left_buffers:
             self.rbufs = self.right_buffers
@@ -146,86 +153,96 @@ class StatsdAggregator(object):
         else:
             self.rbufs = self.left_buffers
             self.wbufs = self.right_buffers
-    
 
-class CounterBucket(object):
-    def __init__(self, name, stat, rate=1, message=None):
+
+class Bucket(object):
+    def format_string(self, udp=False):
+        if udp:
+            return self._to_udp_string()
+        else:
+            return self._to_af_string()
+
+    def _to_udp_string(self):
+        raise NotImplementedError
+
+    def _to_af_string(self):
+        return self._to_udp_string()
+
+
+class CounterBucket(Bucket):
+    def __init__(self, name, stat, rate=1):
         self.name = name
         self.stat = stat
         self.rate = rate
-        self.message = message
 
-    def __str__(self):
-        if self.message:
-            return "%s|c||%s" % (self.stat, self.message)
-        else:
-            return "%s|c" % self.stat
+    def _to_udp_string(self):
+        return "{0}|c".format(self.stat)
 
     def aggregate(self, stat):
-        # Note:  This is non-standard.  We should not divide this out, but instead send the semple rate upstream (with @rate)
+        """
+        CounterBuckets are aggregated by adding new values to
+        the current value.
+        """
+        # Note: This is non-standard. We should not divide this out,
+        #  but instead send the semple rate upstream (with @rate)
         self.stat += int(stat/self.rate)
-        # for chaining
-        return self
 
-class TimerBucket(object):
-    def __init__(self, name, stat, message=None):
+
+class TimerBucket(Bucket):
+    def __init__(self, name, stat):
         self.name = name
-        self.summstat = stat
+        self.stat = [stat]
         self.count = 1
-        self.message = message
 
-    def __str__(self):
-        avg = self.summstat/self.count;
-        if self.message:
-            return "%s|ms||%s" % (avg, self.message)
-        else:
-            return "%s|ms" % avg
+    def _to_af_string(self):
+        """
+        Sending up the full list of values by default so AppFirst can calculate
+        the max/min during the interval as well.
+        """
+        return "{0}|ms".format(','.join([str(n) for n in self.stat]))
+
+    def _to_udp_str(self):
+        """
+        Only send up the average if emitting over UDP so we don't break existing
+        StatsD implementations.
+        """
+        avg = sum(self.stat) / self.count
+        return "{0}|ms".format(avg)
 
     def aggregate(self, stat):
-        self.summstat += stat
+        """
+        TimerBuckets are aggregated by adding new time values to the existing
+        time values and incrementing a counter used to get an average time.
+        """
+        self.stat.extend(stat)
         self.count += 1
-        # for chaining
-        return self
 
-class GaugeBucket(object):
-    def __init__(self, name, stat, message=None):
+
+class GaugeBucket(Bucket):
+    def __init__(self, name, stat):
         self.name = name
         self.stat = stat
-        self.message = ""
         self.timestamp=int(time.time())
 
-    def __str__(self):
-        if self.message:
-            return "%s|g|%s|%s" % (self.stat, self.timestamp, self.message)
-        else:
-            return "%s|g|%s" % (self.stat, self.timestamp)
+    def _to_udp_string(self):
+        return "{0}|g|{1}".format(self.stat, self.timestamp)
 
     def aggregate(self, stat):
+        """
+        GuageBuckets are updated by setting the current gauge value to the new
+        value. No actual aggregation is done.
+        """
         self.stat = stat
-        self.timestamp=int(time.time())
-        # for chaining
-        return self
+        self.timestamp = int(time.time())
 
-#---------------------------------------------------------------------------
-#   Stategy such as where the data is stored and how frequent the stats are sent
-#---------------------------------------------------------------------------
-class GeyserStategy():
-    def __init__(self, interval=10):
-        print "deprecated"
-
-class InstantStategy():
-    def __init__(self):
-        print "deprecated"
 
 #---------------------------------------------------------------------------
 #   Statsd Client
 #---------------------------------------------------------------------------
 class Statsd(object):
-
     _transport = UDPTransport()
     _aggregator = StatsdAggregator(20, _transport)
 
-
     @staticmethod
     def set_transport(transport):
         Statsd._transport.close()
@@ -234,77 +251,70 @@ class Statsd(object):
 
     @staticmethod
     def set_aggregation(should_aggregate):
-        #print "setting aggregation to", should_aggregate
         if should_aggregate and not Statsd._aggregator.running:
             Statsd._aggregator.start()
         if not should_aggregate and Statsd._aggregator.running:
             Statsd._aggregator.stop()
-            
-    @staticmethod
-    def set_strategy(strategy):
-        # deprecated.  Try to fake it
-        if strategy.__name__ == "GeyserStategy":
-            print "Warning: using deprecated method to enable aggregation"
-            Statsd.set_aggregation(True)
-        else:
-            print "Warning: using deprecated method to disable aggregation"
-            Statsd.set_aggregation(False)
 
     @staticmethod
-    def gauge(name, reading, message=None):
+    def gauge(name, reading):
         """
         Log gauge information
         >>> from client import Statsd
         >>> Statsd.gauge('some.gauge', 500)
         """
-        GaugeBucket(name, reading, message)
-        Statsd.send(GaugeBucket(name, reading, message))
+        GaugeBucket(name, reading)
+        Statsd.send(GaugeBucket(name, reading))
 
     @staticmethod
-    def timing(name, elapse, message=None):
+    def timing(name, elapse):
         """
         Log timing information
         >>> from client import Statsd
         >>> Statsd.timing('some.time', 500)
         """
-        Statsd.send(TimerBucket(name, int(round(elapse)), message))
+        Statsd.send(TimerBucket(name, int(round(elapse))))
 
     @staticmethod
-    def increment(names, sample_rate=1, message=None):
+    def increment(names, sample_rate=1):
         """
         Increments one or more stats counters
         >>> Statsd.increment('some.int')
-        >>> Statsd.increment('some.int',0.5)
+        >>> Statsd.increment('some.int', 0.5)
         """
-        Statsd.update_stats(names, 1, sample_rate, message)
+        Statsd.update_stats(names, 1, sample_rate)
 
     @staticmethod
-    def decrement(names, sample_rate=1, message=None):
+    def decrement(names, sample_rate=1):
         """
         Decrements one or more stats counters
         >>> Statsd.decrement('some.int')
         """
-        Statsd.update_stats(names, -1, sample_rate, message)
+        Statsd.update_stats(names, -1, sample_rate)
 
     @staticmethod
-    def update_stats(names, delta=1, sample_rate=1, message=None):
+    def update_stats(names, delta=1, sample_rate=1):
         """
         Updates one or more stats counters by arbitrary amounts
-        >>> Statsd.update_stats('some.int',10)
+        >>> Statsd.update_stats('some.int', 10)
+
+        Sample rate is a decimal value representing the proportion of stats
+        to keep. For example, if sample_rate is 0.5, then 50% of stats will
+        be discarded. Default value is 1 and does not discard any stats.
         """
         if sample_rate < 1 and random.random() > sample_rate:
             return
-        if (type(names) is not list):
+        if not isinstance(names, list):
             names = [names]
         for name in names:
-            Statsd.send(CounterBucket(name, int(round(delta)), sample_rate, message))
+            Statsd.send(CounterBucket(name, int(round(delta)), sample_rate))
 
     @staticmethod
     def send(bucket):
         if Statsd._aggregator.running:
             Statsd._aggregator.add(bucket)
         else:
-            bucket = {bucket.name : bucket}
+            bucket = {bucket.name: bucket}
             Statsd._transport.emit(bucket)
 
     @staticmethod
@@ -313,19 +323,18 @@ class Statsd(object):
 
     @staticmethod
     def shutdown():
-        #print "statsd shutdown"
         Statsd._aggregator.stop()
         Statsd._transport.close()
 
     @staticmethod
     def time(name, enabled=True):
-        '''
+        """
         Function Decorator to report function execution time.
 
-        >>>@Statsd.time("some.timer.bucket")
-        >>>def some_func():
-        >>>    pass #do something
-        '''
+        >>> @Statsd.time("some.timer.bucket")
+        >>> def some_func():
+        >>>     pass  # do something
+        """
         def wrap_timer(method):
             if not enabled:
                 return method
@@ -340,13 +349,13 @@ class Statsd(object):
 
     @staticmethod
     def count(name, sample_rate=1, enabled=True):
-        '''
+        """
         Function Decorator to count how many times a function is invoked.
 
-        @Statsd.count("some.counter.bucket")
-        def some_func():
-            pass #do something
-        '''
+        >>> @Statsd.count("some.counter.bucket")
+        >>> def some_func():
+        >>>     pass #do something
+        """
         def wrap_counter(method):
             if not enabled:
                 return method
@@ -358,17 +367,5 @@ class Statsd(object):
         return wrap_counter
 
 
-'''
-shutdown automatically on application exit...
-'''
-try:
-    import atexit
-    atexit.register(Statsd.shutdown)
-except ImportError: # for Python versions < 2.0
-    def exithook(status, old_exit=sys.exit):
-        try:
-            Statsd.shutdown()
-        finally:
-            old_exit(status)
-
-    sys.exit = exithook
+# shutdown automatically on application exit...
+atexit.register(Statsd.shutdown)
